use core::cmp::min;
use embassy_hal_common::PeripheralRef;
use embassy_embedded_hal::SetConfig;
use embassy_rp::gpio::Level;
use embassy_rp::pio::{Common, Config, FifoJoin, Instance, PioPin, ShiftConfig, ShiftDirection, StateMachine};
use embassy_rp::dma::{Channel};
use embassy_rp::relocate::RelocatedProgram;
use fixed_macro::fixed;
use smart_leds::RGB8;

pub struct Ws2812<'d, P: Instance, D: Channel, const S: usize, const LEDS: usize> {
    sm: StateMachine<'d, P, S>,
    dma: PeripheralRef<'d, D>,
    leds: [u32; LEDS],
}

impl<'d, P: Instance, D: Channel, const S: usize, const LEDS: usize> Ws2812<'d, P, D, S, LEDS> {
    pub fn new(mut pio: Common<'d, P>, mut sm: StateMachine<'d, P, S>, pin: impl PioPin, dma: D) -> Self {
        // Setup sm0

        // prepare the PIO program
        let side_set = pio::SideSet::new(false, 1, false);
        let mut a: pio::Assembler<32> = pio::Assembler::new_with_side_set(side_set);

        const T1: u8 = 2; // start bit
        const T2: u8 = 5; // data bit
        const T3: u8 = 3; // stop bit
        const CYCLES_PER_BIT: u32 = (T1 + T2 + T3) as u32;

        let mut wrap_target = a.label();
        let mut wrap_source = a.label();
        let mut do_zero = a.label();
        a.set_with_side_set(pio::SetDestination::PINDIRS, 1, 0);
        a.bind(&mut wrap_target);
        // Do stop bit
        a.out_with_delay_and_side_set(pio::OutDestination::X, 1, T3 - 1, 0);
        // Do start bit
        a.jmp_with_delay_and_side_set(pio::JmpCondition::XIsZero, &mut do_zero, T1 - 1, 1);
        // Do data bit = 1
        a.jmp_with_delay_and_side_set(pio::JmpCondition::Always, &mut wrap_target, T2 - 1, 1);
        a.bind(&mut do_zero);
        // Do data bit = 0
        a.nop_with_delay_and_side_set(T2 - 1, 0);
        a.bind(&mut wrap_source);

        let prg = a.assemble_with_wrap(wrap_source, wrap_target);
        let mut cfg = Config::default();

        // Pin config
        let out_pin = pio.make_pio_pin(pin);

        let relocated = RelocatedProgram::new(&prg);
        cfg.use_program(&pio.load_program(&relocated), &[&out_pin]);

        // Clock config, measured in kHz to avoid overflows
        // TODO CLOCK_FREQ should come from embassy_rp
        let clock_freq = fixed!(125_000: U24F8);
        let ws2812_freq = fixed!(800: U24F8);
        let bit_freq = ws2812_freq * CYCLES_PER_BIT;
        cfg.clock_divider = clock_freq / bit_freq;

        // FIFO config
        cfg.fifo_join = FifoJoin::TxOnly;
        cfg.shift_out = ShiftConfig {
            auto_fill: true,
            threshold: 24,
            direction: ShiftDirection::Left,
        };

        sm.set_config(&cfg);
        sm.set_pin_dirs(embassy_rp::pio::Direction::Out, &[&out_pin]);
        sm.set_pins(Level::Low, &[&out_pin]);
        sm.set_enable(true);

        let dma = dma.into_ref();
        let leds = [0u32; LEDS];

        Self { sm, dma, leds }
    }

    pub async fn write(&mut self, colors: &[RGB8]) {
        if false {
            for color in colors {
                let word = (u32::from(color.g) << 24) | (u32::from(color.r) << 16) | (u32::from(color.b) << 8);
                self.sm.tx().wait_push(word).await;
            }
        } else {
            for (i, color) in colors.iter().enumerate() {
                if i < LEDS {
                    self.leds[i] = (u32::from(color.g) << 24) | (u32::from(color.r) << 16) | (u32::from(color.b) << 8);
                }
            }
            self.sm.tx().dma_push(self.dma.reborrow(),
                &self.leds[0..min(LEDS, colors.len())]).await;
        }
    }
}
